function app = polydisc(model, cfg)
% Polynomial state, discrete action approximator

CFG.type = 'polydisc';  % # of samples to generate
CFG.disc = 1;
CFG.norm = 0;           % normalize polynomials
CFG.ugrids = [];
CFG.Uflat = [];
CFG.preset = 'lin';     % use a preset; see below for supplied types of presets
CFG.terms = [];         % polynomial terms (powers of each variable on every column)

% These presets are hardcoded for 4 state variables
% All presets include a bias term
% Linear
PRESETS.lin = ...
    [0 0 0 0
     1 0 0 0
     0 1 0 0
     0 0 1 0
     0 0 0 1]';
% Quadratic
PRESETS.quad = ...
    [0 0 0 0
     2 0 0 0
     0 2 0 0
     0 0 2 0
     0 0 0 2]';
% Quadratic + linear terms
PRESETS.quadlin = ...
    [PRESETS.quad'
     1 0 0 0
     0 1 0 0
     0 0 1 0
     0 0 0 1]';
% Quad, linear, and cross-terms
PRESETS.quadcross = ...
    [PRESETS.quadlin'
     1 1 0 0
     1 0 1 0
     1 0 0 1
     0 1 1 0
     0 1 0 1
     0 0 1 1]'; 
PRESETS.quad33 = ...
    [PRESETS.quadcross'
     3 3 0 0        % ^3^3 terms
     0 0 3 3]'; 
 % Quad +
PRESETS.quadplus = ...
    [PRESETS.quadcross'
%      3 1 0 0        % ^3^1 terms
%      1 3 0 0
%      0 0 3 1
%      0 0 1 3
%      3 3 0 0        % ^3^3 terms
%      0 0 3 3
     2 1 0 0      % ^2^1 terms
     2 0 1 0
     2 0 0 1
     1 2 0 0
     0 2 1 0
     0 2 0 1
     1 0 2 0
     0 1 2 0
     0 0 2 1
     1 0 0 2
     0 1 0 2
     0 0 1 2
     ]'; 
% Full cubic
PRESETS.cube = ...
    [PRESETS.quadcross'
     3 0 0 0
     0 3 0 0
     0 0 3 0
     0 0 0 3
     1 1 1 0
     1 1 0 1
     1 0 1 1
     0 1 1 1
     2 1 0 0
     2 0 1 0
     2 0 0 1
     1 2 0 0
     0 2 1 0
     0 2 0 1
     1 0 2 0
     0 1 2 0
     0 0 2 1
     1 0 0 2
     0 1 0 2
     0 0 1 2
     ]'; 
% derived from Lagoudakis and Parr bicycle riding (minus the orientation variable)
PRESETS.lp = ...
    [0 0 0 0
     1 0 0 0
     0 1 0 0
     2 0 0 0
     0 2 0 0
     1 1 0 0
     0 0 1 0
     0 0 0 1
     0 0 2 0
     0 0 0 2
     0 0 1 1
     1 0 1 0
     1 0 2 0
     2 0 1 0]';  
 
cfg = parseconfig(cfg, CFG);
if isempty(cfg.terms),
    cfg.terms = PRESETS.(cfg.preset);
end;

app = copyfields(cfg, struct, {'type', 'terms', 'norm', 'preset'});
% core functions
if ~cfg.norm,   % not normalized
    app.phi = @polydisc_phi; app.q = @polydisc_q; app.h = @polydisc_h; 
else            % normalized
    app.phi = @polydisc_phin; app.q = @polydisc_qn; app.h = @polydisc_hn; 
end;    
% plot functions
app.plotv = @polydisc_plotv; app.ploth = @polydisc_ploth;

app.maxx = model.maxx;
% terms
app.Nt = size(app.terms, 2);
% action space, structured and flat
if ~isempty(cfg.ugrids),
    app.U = cfg.ugrids; 
    app.Uflat = flat(cfg.ugrids);
elseif ~isempty(cfg.Uflat),
    app.Uflat = cfg.Uflat;
    app.U = unflat(app.Uflat); 
else
    error('Discrete actions could not be obtained');
end;
% number of actions
app.M = size(app.Uflat, 2);
% finally, total number of basis functions
app.N = app.Nt * app.M;

end     % constructor

% =====================
% UNNORMALIZED versions
function phi = polydisc_phi(app, x, u)
phi = zeros(app.N, 1);
ui = findflat(u, app.Uflat, 1, 'first');
phi( ((ui-1)*app.Nt)+1 : (ui*app.Nt) ) = prod(repmat(x, 1, app.Nt) .^ app.terms);
end
%
function q = polydisc_q(app, theta, x, u)
ui = findflat(u, app.Uflat, 1, 'first');
q = prod(repmat(x, 1, app.Nt) .^ app.terms) * theta( ((ui-1)*app.Nt)+1 : (ui*app.Nt) );
end
%
function u = polydisc_h(app, theta, x)
qx = prod(repmat(x, 1, app.Nt) .^ app.terms) * reshape(theta, app.Nt, app.M);
u = app.Uflat(:, find(qx == max(qx), 1, 'first'));
end

% =====================
% NORMALIZED versions
% Note this type of normalization is incorrect because polys can also take negative values!!!
% function phi = polydisc_phin(app, x, u)
% phi = zeros(app.N, 1);
% ui = findflat(u, app.Uflat, 1, 'first');
% phix = prod(repmat(x, 1, app.Nt) .^ app.terms);
% phi( ((ui-1)*app.Nt)+1 : (ui*app.Nt) ) = phix ./ sum(phix);
% end
% % 
% function q = polydisc_qn(app, theta, x, u)
% ui = findflat(u, app.Uflat, 1, 'first');
% phix = prod(repmat(x, 1, app.Nt) .^ app.terms);
% q = (phix ./ sum(phix)) * theta( ((ui-1)*app.Nt)+1 : (ui*app.Nt) );
% end
% %
% function u = polydisc_hn(app, theta, x)
% phix = prod(repmat(x, 1, app.Nt) .^ app.terms);
% qx = (phix ./ sum(phix)) * reshape(theta, app.Nt, app.M);
% u = app.Uflat(:, find(qx == max(qx), 1, 'first'));
% end

% ----- Plot functions
% Plot V - forward call to generic plot fun
function polydisc_plotv(app, theta, varargin)
for i = length(app.maxx):-1:1, xb(i, 1:2) = [-app.maxx(i) app.maxx(i)]; end;
for i = length(app.U):-1:1, ub(i, 1:2) = app.U{i}([1 end]); end;
approx_plotv(app, theta, xb, ub, varargin{:});
end
% Plot h - forward call to generic plot fun
function polydisc_ploth(app, theta, varargin)
for i = length(app.maxx):-1:1, xb(i, 1:2) = [-app.maxx(i) app.maxx(i)]; end;
for i = length(app.U):-1:1, ub(i, 1:2) = app.U{i}([1 end]); end;
approx_ploth(app, theta, xb, ub, varargin{:});
end
