function varargout = ersarsa(cfg)
% Q-learning
%   VARARGOUT = ERSARSA(CFG)
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see he str2cfg
% Outputs:
%   Q               - in run mode. Computed Q-function.
%   HIST, FIGH      - in replay mode. HIST is the replay history.
%           FIGH contains figure handles if figures were created 
%           and not closed; and is an empty matrix if all the figures were closed.

% NOTE THIS COMES FROM SANDER AND WAS NOT CHECKED /EDITED

if nargin < 1, cfg = struct(); end;

% -----------------------------------------------
% Process configuration structure

% default config
% script config
CFG.model_params = {};              % extra parameters for problem calling in 'model' mode
CFG.run = 0;                        % run learning
CFG.replay = 0;                     % replay learned policy
CFG.problem = '';                   % what problem to solve
CFG.datadir = '';                   % data dir
CFG.datafile = '';                  % save data to file
CFG.gamma = 0.98;                   % discount factor
CFG.alpha = .1;                     % learning rate
CFG.explor = .9;                    % exploration rate
CFG.explordecay = .9;               % exploration decay (per trial)
CFG.lambda = 0;                     % eligibility trace parameter
CFG.eps = .01;                      % threshold for convergence
CFG.maxtrials = 100;                % max # trials
CFG.maxsteps = 1000;                % max # steps in a trial
CFG.reset = 'rand';                 % trial reset
CFG.randseed = [];                  % set the random seed (for repeatable results)
% replay config
CFG.x0 = [];                        % initial state for replay (otherwise the problem default or zeros)
CFG.tend = 30;                      % end time for replay
% experience replay config
CFG.er_N = 10;                      % experience replay number of replays N
CFG.er_level = 'samples';           % level of replay, single samples or whole trajects
CFG.er_chose_sample = '';
CFG.er_chose_length1 = '';
CFG.er_chose_length2 = '';
CFG.plottarget = 'screen';          % 'screen', '', 'latex', or 'beamer'. If 'screen' figures will not be closed
CFG.savedir = '';
CFG.savefig = '';
% display config
CFG.verb = 5;                       % verbosity: the higher, the more detailed the messages displayed
CFG.visualize = 0;                  % visualization level (0 = none, 1 = trial-level, 2 = step-level)
CFG.viscfg = struct;                % visualization config options
CFG.trialdisp = 10;

% List of fields that define the problem and therefore may NOT be overwritten on load
KEEPFIELDS = {'problem', 'gamma'};

% Install function defaults for everything else
cfg = parseconfig(cfg, CFG);

cfg.init = cfg.run || ~exist([cfg.datafile '.mat'], 'file');

if ~cfg.init,        % load data file, making sure that cfg is not overwritten
    cfg1 = cfg; kf = KEEPFIELDS;
    load(cfg.datafile);
    % Overwrite problem-defining fields from loaded config, keep the rest as in the (current) cfg1;
    % the result becomes the new config
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['Data loaded from [' cfg.datafile '].'], cfg.verb, 1);
end;

% Echo config
dispx('Q-learning called with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1);

% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

% -----------------------------------------------
% Create model, find dimensionality data
if cfg.init,
    model = feval(cfg.problem, 'model', cfg.model_params{:});
    DIMS.dimx = zeros(model.p, 1);
    DIMS.dimu = zeros(model.q, 1);
    for i = 1:model.p, DIMS.dimx(i) = length(model.X{i}); end;
    for i = 1:model.q, DIMS.dimu(i) = length(model.U{i}); end;
    DIMS.N = prod(DIMS.dimx);
    DIMS.M = prod(DIMS.dimu);
    Xflat = flat(model.X); Uflat = flat(model.U);
    p = model.p; q = model.q;
end;

% -----------------------------------------------
% Q-iteration
if cfg.run,
    
    % init Q-function, elig trace, histories etc.
    X = NaN(p, cfg.maxsteps+1, cfg.maxtrials);
    U = NaN(q, cfg.maxsteps+1, cfg.maxtrials);
    R = NaN(1, cfg.maxsteps+1, cfg.maxtrials);
    Q = zeros(DIMS.N, DIMS.M);
    E = zeros(DIMS.N, DIMS.M);      % not yet used
    Qh = cell(cfg.maxtrials+1, 1);  
    Qh{1} = Q;                      % save Q_0 on the stats
    deltah = NaN(cfg.maxtrials+1, 1);
    
    switch cfg.er_level
        case 'samples'              % replay N single random samples
            cfg.er_chose_sample = 'randi(er_trial_length-1)';
            cfg.er_chose_length1 = 'cfg.er_N*avg_trial_length';
                                    %N1 = num_samples; % let all samples have a chance to be replayed
            cfg.er_chose_length2 = '0';
        case 'trajectories'         % or N random trajectories as a whole
            cfg.er_chose_sample = 'rk + 1';
            cfg.er_chose_length1 = 'cfg.er_N';
            cfg.er_chose_length2 = 'er_trial_length - 2';
        case 'trajectories_back'    % or N random trajectories as a whole, but backward
            cfg.er_chose_sample = 'er_trial_length - rk - 1';
            cfg.er_chose_length1 = 'cfg.er_N';
            cfg.er_chose_length2 = 'er_trial_length - 2';
        otherwise
            error('Non-supported level for Experience Replay, use samples, trajectories or trajectories_back');
    end;

    trial = 1;
    timestat.run = 0;
    explor = cfg.explor;
    alpha = cfg.alpha;
    
    if ~isempty(cfg.randseed), rand('twister', cfg.randseed); end;
    
    % init visualization config if needed
    if cfg.visualize,
        vcfg = cfg.viscfg;
        vcfg.gview = [];
    end;

    % -----------------------------------------------
    % Perform Q-iteration
    if cfg.lambda > 0, dispx('Performing ER-SARSA-learning...', cfg.verb, 0);
    else dispx('Performing ER-SARSA...', cfg.verb, 0);
    end;

    t = cputime;
    conv = 0;
    while trial <= cfg.maxtrials && ~conv,       % main loop

        % initialize trial
        k = 1; terminal = 0;
        % initial state
        if isnumeric(cfg.reset),            
            if size(cfg.reset, 2) > 1,      X(:, 1, trial) = cfg.reset(:, unidrnd(size(cfg.reset, 2)));
            else                            X(:, 1, trial) = cfg.reset;
            end;
        % note a 'rand' reset to a terminal state will cause the algorithm
        % to signal convergence since nothing is learned during the trial!
        elseif strcmp(cfg.reset, 'rand'),   X(:, 1, trial) = Xflat(:, unidrnd(DIMS.N));
        end;
        xind = findflat(X(:, 1, trial), Xflat, 1, 'first');
        % reset elig trace
        E = 0 .* E;
        % take first action
        if rand < explor,   % exploratory
            uind = unidrnd(DIMS.M);
        else                % greedy
            [Qmax uind] = find(Q(xind, :) == max(Q(xind, :)));
            if length(uind) > 1, uind = uind(unidrnd(length(uind))); end;
        end;
        U(:, 1, trial) = Uflat(:, uind);    % transform into actual action

        timestat.run = timestat.run + (cputime - t);
        % visualize initial state
        if cfg.visualize >= 2,
            vcfg.ersarsastep = 1;
            vcfg.ersarsatrial = 0;
            vcfg.ersarsarep = 0;
            vcfg.k = k - 1;
            vcfg.trial = trial;
            [figh vcfg.gview] = feval(model.visualizefun, vcfg);
        end;
        t = cputime;
        
        % do the steps    
        while k < cfg.maxsteps && ~terminal,
            
            % decay trace
            if cfg.lambda > 0, E = (cfg.gamma * cfg.lambda) .* E; end;

            % apply chosen action
            [X(:, k+1, trial), R(1, k+1, trial), terminal] = ...
                feval(model.fun, model, X(:, k, trial), U(:, k, trial));
            xindplus = findflat(X(:, k+1, trial), Xflat, 1, 'first');
            
            % choose action in the NEXT state (required to compute the
            % temporal difference)
            if rand < explor,   % exploratory
                uindplus = unidrnd(DIMS.M);
            else                % greedy action
                [Qmax uindplus] = find(Q(xindplus, :) == max(Q(xindplus, :)));
                if length(uindplus) > 1, uindplus = uindplus(unidrnd(length(uindplus))); end;
            end;
            % transform into the actual action
            U(:, k+1, trial) = Uflat(:, uindplus);
            
            % compute temporal difference and update Q-function
            tempdiff = R(1, k+1, trial) + cfg.gamma * Q(xindplus, uindplus) - Q(xind, uind);
            if cfg.lambda > 0,  % eligibility trace based update
                E(xind, uind) = 1;                  % replace trace of crt state-action pair
                Q = Q + (alpha * tempdiff) .* E;    % update Q-function
            else                % quick, single-element update
                Q(xind, uind) = Q(xind, uind) + alpha * tempdiff;        
            end;

            % update stats
            timestat.run = timestat.run + (cputime - t);
            
            % visualization
            if cfg.visualize >= 2,
                vcfg.ersarsastep = 1;
                vcfg.ersarsatrial = 0;
                vcfg.ersarsarep = 0;
                vcfg.k = k;
                vcfg.trial = trial;
                [figh vcfg.gview] = feval(model.visualizefun, vcfg);
            end;
            
            % (in the future, if alpha must be annealed, it should be done here)
            
            % Experience replay
            
            if trial > 1 % can not replay trials if this is the first trial
                num_samples = sum(sum(~isnan(X(1, :, :))));
                avg_trial_length = num_samples/trial;
                er_i = 1;

                N1 = eval(cfg.er_chose_length1);
                for rt=1:N1
                    er_trial = randi(trial-1);
                    er_trial_length = sum(~isnan(X(:, :, er_trial)));
                    N2 = eval(cfg.er_chose_length2);

                    for rk=0:N2
                        er_sample = eval(cfg.er_chose_sample);
                        % retreive sample from database
                        xrepind = findflat(X(:, er_sample, er_trial), Xflat, 1, 'first');
                        xrepindplus = findflat(X(:, er_sample+1, er_trial), Xflat, 1, 'first');
                        urepind = findflat(U(:, er_sample, er_trial), Uflat, 1, 'first');
                        urepindplus = findflat(U(:, er_sample+1, er_trial), Uflat, 1, 'first');
                        rrep = R(:, er_sample+1, er_trial);
                        
                        % compute temporal difference and update Q-function
                        tempdiff = rrep + cfg.gamma * Q(xrepindplus, urepindplus) - Q(xrepind, urepind);
                        Q(xrepind, urepind) = Q(xrepind, urepind) + alpha * tempdiff;
                        
                        % visualization
                        if cfg.visualize >= 2,
                            vcfg.ersarsastep = 0;
                            vcfg.ersarsatrial = 0;
                            vcfg.ersarsarep = 1;
                            vcfg.k = k;
                            vcfg.trial = trial;
                            vcfg.er_i = er_i;
                            vcfg.er_k = er_sample;
                            vcfg.er_trial = er_trial;
                            [figh vcfg.gview] = feval(model.visualizefun, vcfg);
                        end;

                        er_i = er_i + 1;
                    end;
                end;
            end;
            
            t = cputime; 
            k = k + 1;
            xind = xindplus;
            uind = uindplus;
        end;
            
        % store Q-function on history
        Qh{trial+1} = Q;
        % compute max absolute difference
        deltah(trial+1) = max(max(abs(Q - Qh{trial})));
        conv = deltah(trial+1) < cfg.eps;

        % update stats
        timestat.run = timestat.run + (cputime - t);
        
        % trial visualization
        if cfg.visualize >= 1,
            vcfg.ersarsastep = 0;
            vcfg.ersarsatrial = 1;
            vcfg.ersarsarep = 0;
            vcfg.trial = trial;
            [figh vcfg.gview] = feval(model.visualizefun, vcfg);
        end;
        
        % console feedback of algorithm progress
        if ~mod(trial, cfg.trialdisp), 
            dispx(sprintf('Trial #%d completed, delta=%.3f', trial, deltah(trial+1)), cfg.verb, 2);
        end;
        
        % start counting time again, increment iteration counter
        t = cputime;
        trial = trial + 1;
        % anneal exploration
        explor = explor * cfg.explordecay;
    end;        % while not converged and allowed more iterations

    if conv,	dispx('Convergence detected. Algorithm stopped.', cfg.verb, 0);
    else        dispx(sprintf('maxtrials=%d exhausted. Algorithm stopped.', cfg.maxtrials), cfg.verb, 0);
    end;
    
    % finalize visualizer
    if cfg.visualize,
        vcfg.ersarsastep = 0;
        vcfg.ersarsatrial = 0;
        vcfg.ersarsarep = 0;
        vcfg.finalize = 1;
        [figh vcfg.gview] = feval(model.visualizefun, vcfg);
    end;
    
    % output Q-function
    varargout = {Q};
end;

% -----------------------------------------------
% Backup data

if cfg.run && ~isempty(cfg.datafile),
    % save into the same directory as the problem
    if isempty(cfg.datadir),
        datadir = fileparts(which(cfg.problem));
    else
        datadir = cfg.datadir;
        if (datadir(end) == '\'), datadir = datadir(1:end-1); end;
    end;
    cfg.datafile = [datadir '\' cfg.datafile];
    save(cfg.datafile);
    dispx(['Data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;


% END qi() RETURNING varargout =================================================================
