function [eJ, J] = mc_rbfqi(varargin)
% Easy-to-use wrapper for mc_rbfq
%  [EJ, J, K] = MC_RBFQI(CFG)
%  [EJ, J, K] = MC_RBFQI(DATAFILE, X0)
%  [EJ, J, K] = MC_RBFQI(DATAFILE, X0, K)
%  [EJ, J, K] = MC_RBFQI(MODEL, GAMMA, C, RAD, THETA, X0, U, K)
%
% Parameters or configuration fields:
%   DATAFILE    - from where to load the fuzzy Qiteration results
%   X0          - the set of states for which to compute the return
%   K           - maximum length of a simulation
%   GAMMA       - the discount factor
%   C           - centers
%   RAD         - radii
%   THETA       - parameter matrix
%   U           - discrete action space
% For the rest of the configuration fields, see the commented defaults in
% the code
% Returns:
%   EJ          - mean return over X0
%   J           - cost for every state in X0
%   K           - number of stmc_eps until termination, otherwise K, for every
%           state in X0

CFG.datafile = [];
CFG.model = [];
CFG.gamma = [];
CFG.c = [];
CFG.rad = [];
CFG.theta = [];
CFG.X0 = [];
CFG.U = [];
CFG.K = 200;
CFG.mc_eps = [];        % precision in return eval. If set, supersedes the K setting
CFG.userew = '';        % use this reward function instead of what the algorithm was run with

if nargin == 1,
    cfg = varargin{1};
    if ischar(cfg), cfg = str2cfg(cfg, fieldnames(CFG)); end;
elseif nargin == 2,
    cfg.datafile = varargin{1};
    cfg.X0 = varargin{2};
elseif nargin == 3,
    cfg.datafile = varargin{1};
    cfg.X0 = varargin{2};
    cfg.K = varargin{3};
elseif nargin == 7,
    cfg.model = varargin{1};
    cfg.gamma = varargin{2};
    cfg.c = varargin{3};
    cfg.rad = varargin{4};
    cfg.theta = varargin{5};
    cfg.X0 = varargin{6};
    cfg.U = varargin{7};
    cfg.K = varargin{8};
end;
cfg = checkparams(cfg, CFG);

if ~isempty(cfg.datafile),
    loadwithprefix(cfg.datafile, 'd_', {'cfg', 'model', 'c', 'rad', 'theta', 'U'});
    cfg.model = d_model;
    cfg.gamma = d_cfg.gamma;
    cfg.c = d_c;
    cfg.rad = d_rad;
    cfg.theta = d_theta;
    cfg.U = d_U;
    if isempty(cfg.X0), cfg.X0 = unflat(d_c); end;
end;

% change reward if needed
if ~isempty(cfg.userew),
    cfg.model = feval(d_cfg.problem, 'changerew', cfg.model, cfg.userew);
end;

% mc_eps takes precedence over K
if ~isempty(cfg.mc_eps),
    cfg.K = ceil(log(cfg.mc_eps * (1-cfg.gamma) / cfg.model.maxr) / log(cfg.gamma));    
end;

mccfg.U = cfg.U;
mccfg.mc_maxsteps = cfg.K;
mccfg.gamma = cfg.gamma;
if iscell(cfg.X0),
    for i = length(cfg.X0):-1:1,
        N0(i) = length(cfg.X0{i});
    end;
    cfg.X0 = flat(cfg.X0);
else N0 = [];
end;

% forward the call to mc_rbfq
[eJ, J] = mc_rbfq(cfg.c, cfg.rad, cfg.theta, cfg.model, cfg.X0, mccfg);

if ~isempty(N0),
    J = reshape(J, N0);
end;
