function varargout = lspi_extractsamples(cfg)
% Extract samples of Q, V, or h from a LSPI solution
% See below for commented configuration (input) arguments
% The output is:
%   - in Q mode: [Q, X, U, Nx, Nu]
%   - in V mode: [V, X, Nx]
%   - in h mode: [h, X, Nx]
% where:
%   - Q is an array of Q-values of dimension [Nx Nu] 
%   - V is an array of state values of dimension Nx
%   - h is an array of actions of dimension Nx
%   - X is a cell vector of state grids, one grid for each state dimension. The vector has
%   size p where p is the # of states in the system
%   - Nx contains the dimensions of each grid in X, in a vector of size p
%   - U is a cell vector of action grids, one grid for each action dimension. The vector has
%   size p where q is the # of actions in the system
%   - Nu contains the dimensions of each grid in U, in a vector of size q
%
% Note that in Q mode for a discrete-action problem, U and Nu will contain the action
% discretization and its dimension, and cannot be controlled by the user.
%
% The syntax is common for discrete- or continuous- action problems.

% Defaults
% what to compute (the first of these takes priority, two objects cannot be computed with the
% same function call)
CFG.Q = 0;
CFG.V = 0;
CFG.h = 0;
% data file from where to load the approximator, the model, and the LSPI solution
CFG.datafile = '';
% Number of equidistant points to sample
CFG.Nx = 20;    % number of points on each axis of X in Q, V, h; 
    % if a scalar, the same number will be used for every axis
CFG.Nu = 10;    % number of points on each axis of U in Q; ignored for V, h, and discrete-actions
    % if a scalar, the same number will be used for every axis
    
cfg = parseconfig(cfg, CFG);

load(cfg.datafile, 'model', 'approx', 'thetastar');

X = cell(1, model.p);
if isscalar(cfg.Nx), cfg.Nx = cfg.Nx + zeros(1, model.p); end;
for i = 1:model.p,
    X{i} = -model.maxx(i) : 2*model.maxx(i)/(cfg.Nx(i)-1) : model.maxx(i);
end;
Xf = flat(X);
Nxf = prod(cfg.Nx);

if cfg.Q, 
    if any(strcmp(approx.type, {'triangquad', 'rbfquad'})),
        disp('Computing continuous-action Q on an X-U grid. Returning Q, X, U, Nx, Nu');
        U = cell(1, model.q);
        if isscalar(cfg.Nu), cfg.Nu = cfg.Nu + zeros(1, model.q); end;
        for i = 1:model.q,
            U{i} = -model.maxu(i) : 2*model.maxu(i)/(cfg.Nu(i)-1) : model.maxu(i);
        end;
        Uf = flat(U);
        Nuf = prod(cfg.Nu);
        Q = zeros(Nxf, Nuf);
        for i = 1:Nxf,
            for j = 1:Nuf,
                Q(i, j) = approx.q(approx, thetastar, Xf(:, i), Uf(:, j));
            end;
        end;
        Q = reshape(Q, [cfg.Nx cfg.Nu]);
        varargout = {Q, X, U, cfg.Nx, cfg.Nu};
    else 
        disp('Computing discrete-action Q on an X grid. Returning Q, X, U, Nx, Nu');
        Q = zeros(Nxf, approx.M);
        for i = 1:Nxf,
            for j = 1:approx.M,
                Q(i, j) = approx.q(approx, thetastar, Xf(:, i), approx.Uflat(:, j)); 
            end;
        end;
        Q = reshape(Q, [cfg.Nx approx.M]);
        [U, Nu] = unflat(approx.Uflat);
        varargout = {Q, X, U, cfg.Nx, Nu};
    end;
elseif cfg.V,
	disp('Computing V on an X grid. Returning V, X, Nx');
    V = zeros(Nxf, 1);
    for i = 1:Nxf,
        ustar = approx.h(approx, thetastar, Xf(:, i)); 
        V(i) = approx.q(approx, thetastar, Xf(:, i), ustar); 
    end;
    V = reshape(V, cfg.Nx);
    varargout = {V, X, cfg.Nx};
elseif cfg.h,
	disp('Computing h on an X grid. Returning h, X, Nx');
    h = zeros(Nxf, 1);
    for i = 1:Nxf,
        h(i) = approx.h(approx, thetastar, Xf(:, i)); 
    end;
    h = reshape(h, cfg.Nx);
    varargout = {h, X, cfg.Nx};
end;
