function  [uroot, planinfo, plandata] = olop_theoretical(xroot, model, cfg)
% OLOP Open-loop optimistic planning, theoretical variant
%   Detailed explanation goes here

% Author: Lucian Busoniu
% Limitation: no handling of terminal states
% NOTE: first discount is gamma^1 here, not gamma^0

CFG.gamma = [];             % discount factor
CFG.n = [];                 % simulation budget
CFG.U = [];                 % discrete, flat action space
CFG.k = -1;                 % discrete time step
% display config
CFG.verb = 0;               % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
U = cfg.U;
M = size(U, 2);      % number of discrete actions

% start counting time
t = cputime;

% find by trial and error the largest # simulations M so that M ceil[log M / (2 log 1/gamma)] <= n
T = 1; while T * ceil(log(T) / (2*log(1/cfg.gamma))) <= cfg.n, T = T+1; end; T = T-1;
% corresponding history length
H = ceil( log(T) / (2*log(1/cfg.gamma)) );              

% init sub/sequence information
SEQ = zeros(H, 0, 'uint8');         % sequences (uint8 as M > 256 is unfeasible)
Hseq = zeros(1, 0, 'uint8');        % depths (horizons deeper than 256 unfeasible)
Cseq = zeros(M, 0, 'uint32');       % indices of children
Tseq = zeros(1, 0, 'double');       % counters (kept double as they are used in math)
Rseq = zeros(1, 0, 'double');       % rewards
UCBseq = Rseq;                      % UCBs
Bseq = Rseq;                        % B-values
% simulation history
Xh = nan(model.p, H+1, T); Uh = nan(model.q, H+1, T); Rh = nan(1, H+1, T);

if cfg.verb >= 4, disp(sprintf('Starting planning, T=%d, H=%d, n=%d out of %d', ...
        T, H, T*H, cfg.n)); end;
            
% main simulation loop
discount = cfg.gamma.^(1:H+1);             % note starting from gamma^1, not gamma^0
zerosH = zeros(H, 1);
for tau = 1:T,
    
    % choose a sequence that maximizes B
    % note when several choices are equally good, we go for the first for more repeatable behavior
    if tau <= M,     
        % not all of the initial actions have been played, so there are still branches of the tree
        % with "infinity" (i.e. undefined) B values -- choose the first, fill with ones up to H
        % (the test assumes we take the initial actions in order)
        seq = [tau; ones(H-1, 1)];
    else
        % all nodes have finite B-values, choose the maximum
        % note that shorter subsequences represent entire subtrees, and any sequences in those
        % substrees are eligible; so we select any such (sub)sequence and fill with ones up to H
        [Bmax, imax] = max(Bseq);
        seq = SEQ(:, imax);
        seq(~seq) = 1;
    end;
    
    % apply the sequence of actions, starting from xroot
    Xh(:, 1, tau) = xroot;
    for h = 1:H,
        Uh(:, h, tau) = U(:, seq(h));
        [Xh(:, h+1, tau), Rh(1, h+1, tau), term] = feval(model.fun, model, Xh(:, h, tau), Uh(:, h, tau));
        if term, error('OLOP does not currently support terminal states'); end;
    end;
    
    % process subsequences up to each h
    sub = zerosH; isub = [];
    for h = 1:H,
        % with this, sub = [seq(1:h) 0 ... 0]
        sub(h) = seq(h);
        % find the subsequence or otherwise create it
        i = findflat(sub, SEQ);
        if isempty(i),
            i = size(SEQ, 2) + 1;
            SEQ(:, i) = sub;
            Tseq(i) = 0;
            Rseq(i) = 0;
            Hseq(i) = h;
            % this is the m-th (sub(h)th) child for the parent subsequence (isub(h-1))
            if h > 1, Cseq(sub(h), isub(h-1)) = i; end;
        end;
        isub(h) = i;
        % increment counter, add obtained reward
        Tseq(i) = Tseq(i) + 1;
        Rseq(i) = Rseq(i) + Rh(1, h+1, tau);
        % set (or update, if existing) UCB
        UCBseq(i) = sum( discount(1:h) .* ...
                ( Rseq(isub) ./ Tseq(isub)  +  sqrt(2*log(T) ./ Tseq(isub)) ) ...
            ) + discount(h+1) / (1-cfg.gamma);
    end;
    
    % update B-values for all existing sequences:
    % iterate forward through the equivalent tree, setting at each stage 
    % the (initialized) childrens' values to minimum of their parent's and their existing values
    Bseq = UCBseq;
    for h = 1:H-1,
        ih = find(Hseq == h);
        for i = ih,
            children = Cseq(Cseq(:, i) > 0, i);
            Bseq(children) = min(Bseq(i), Bseq(children));
        end;
    end;

    % log
    if cfg.verb >= 5, disp(sprintf('Trajectory tau=%d processed', tau)); end;
    if cfg.verb >= 6, 
        % show detailed stats
        disp('U/R'); disp([Uh(:, :, tau); Rh(1, :, tau)]);
        disp('Counters, rewards, UCBs, B-values for initialized sequences');
        S = cell(size(SEQ, 2), 1);
        spec = ['%-' num2str(H) 's: T=%3d, R=%2.2f, UCB=%2.2f, B=%2.2f'];
        for i = 1:size(SEQ, 2),
            S{i} = sprintf(spec, char(double(SEQ(SEQ(:, i) > 0, i))+'0'), Tseq(i), Rseq(i), UCBseq(i), Bseq(i));
        end;
        disp(S);
        pause;
    end;
        
end;    % end main loop

% finally, return an action that has been played the most times
[tmax, imax] = max(Tseq);
uroot = U(:, SEQ(1, imax));

% output stats if requested
if nargout >= 2,
    planinfo.T = T;
    planinfo.H = H;
    planinfo.n = T*H;
    planinfo.trun = cputime - t;
    planinfo.message = sprintf('Planning done, T=%d traj, horizon H=%d, n_eff=%d steps', ...
        T, H, planinfo.n); 
    if cfg.verb >= 4, disp(planinfo.message); end;
end;
if nargout >= 3,
    plandata = varstostruct('SEQ', 'Tseq', 'Rseq', 'UCBseq', 'Bseq', 'Xh', 'Uh', 'Rh');
end;

end % main function

