function  [uroot, planinfo, tree] = opss_objects(xroot, model, cfg)
% OPSS Optimistic planning for sparsely stochastic systems
%   New version using handle objects for the nodes
%   Still buggy - not bug-checked, runs very slowly

% Author: Lucian Busoniu
% Versions:

CFG.gamma = [];             % discount factor
CFG.n = [];                 % expansion budget
CFG.criterion = 'contrib';  % expansion criterion: based on the whole tree: 'unif'
                            % based on optimistic subtree: 'contrib'; for other modes see
                            % opss_compatibility
CFG.actsel = 'util';        % action selection at root: 'util'; for other modes see opss_compatibility
CFG.L = [];                 % maximum number of successors for each state-action pair
CFG.U = [];                 % discrete, flat action space
CFG.reusetree = 0;          % try reusing previous tree (only when k is set and > 1)
CFG.maxtreesize = 50000;    % do not reuse tree if larger (has more nodes) than this
CFG.k = -1;                 % discrete time step
CFG.compat = 0;             % force using compatibility version
% display config
CFG.verb = 0;               % verbosity: the higher, the more detailed the messages displayed

% persistent data;
cfg = checkparams(cfg, CFG);

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.gamma;
structtovars(cfg, {'n', 'criterion', 'L', 'U'});
M = size(U, 2);      % number of discrete actions

% start counting time
t = cputime;

% prepare empty node template
tree = [];
if isempty(tree),
    % initialize the tree to contain a single (root) node corresponding to xroot
    root = opssnode;
    root.parent = [];
    root.parentj = [];
    root.children = [];
    root.depth = 0;
    root.leaf = true;
    root.x = xroot;
    root.f = 1;
    root.r = 0;
    root.term = false;   % presumably, we are not asked to make a decision in a terminal state
    root.b = repmat(1/(1-discount), M, 1);
    root.util = zeros(M, 1);
    root.P = 1;
    root.R = 0;
    
    tree = root;
end;

% main tree construction loop
for tau = 1:n,
    
    % 1) find the candidate leaves (frontier) for expansion
    if criterion(1) == 'c',
        % greedy leaves (forward traversal)
        frontier = root;  % init with root
        while ~all([frontier.leaf]),         % not all leaves
            nextnodes = [];                  % new nodes to add
            for node = frontier(~[frontier.leaf]),     % loop through each inner node
                % find a greedy branch, breaking ties deterministically (i.e. choose first greedy branch)
                [~, jmax] = max(node.b);
                % add children on greedy branch
                nextnodes = [nextnodes node.children{jmax}];
            end; 
            % new set of nodes is old set of LEAF nodes, plus the newly added nodes irrespective of type
            frontier = [frontier([frontier.leaf]) nextnodes];
        end;
    elseif criterion(1) == 'u',     % uniform expansion -- just get all the leaves
        frontier = tree([tree.leaf]);
    end;
    
    % 2) select a leaf node to expand
    % we know all nodes are leaves, but some of them may be terminal
    % eliminate the terminal nodes and continue with the rest
    frontier = frontier(~[frontier.term]);
    % if ALL are terminal, there is nothing left to do
    if isempty(frontier), break; end;
    if criterion(1) == 'c',
        % largest contribution to the bound (breaking ties determ)
        % note we do not consider the factor 1/(1-gamma) as it appears everywhere
        % this is the variant used in the ADPRL paper
        C = [frontier.P] .* discount .^ [frontier.depth];
        [~, imax] = max(C);
        nodetoexpand = frontier(imax);
    elseif criterion(1) == 'u',     
        % uniform: expand a node with the smallest depth (breaking ties determ)
        [~, imin] = min([frontier.depth]);
        nodetoexpand = frontier(imin);
    end;
    
    % 3) expand the chosen node
    nodetoexpand.leaf = false;                % it will no longer be a leaf
    nodetoexpand.children = cell(M, 1);       % init children table
    % run through the discrete actions 
    for j = 1:M,
        % use model to determine distribution over next states (+ rewards, terminal flags)
        [x, r, term, f] = feval(model.fun, model, nodetoexpand.x, U(:, j));
        if any(r < 0) || any(r > 1), error('Rewards must be in [0, 1]'); end;
        % note we allow FEWER than L actions here
        if length(f) > L, error('Too many successors (%d, at most %d allowed)', length(f), L); end;
        % run through the possible next states
        for ell = 1:length(f),
            % create new node
            child = opssnode;
            % meta-info
            child.parent = nodetoexpand;
            child.parentj = j;
            child.children = [];
            child.depth = nodetoexpand.depth + 1;
            child.leaf = true;
            % data
            child.x = x(:, ell);
            child.f = f(ell);
            child.r = r(ell);
            child.term = term(ell);
            if term(ell),   child.b = zeros(M, 1);
            else            child.b = repmat(1/(1-discount), M, 1);
            end;
            child.util = zeros(M, 1);
            child.P = nodetoexpand.P * f(ell);
            child.R = nodetoexpand.R + discount^nodetoexpand.depth * r(ell);
            
            % save child in parent children table
            nodetoexpand.children{j}(ell) = child;
            % and on the flat tree structure
            tree(end+1) = child;
        end;
    end;
    
    % 4) update b-values affected by expanded node
    % these are the b-values of the expanded node itself for all the actions, 
    % and those on the state-ACTION path from the expanded added node to the root
    node = nodetoexpand; jrange = 1:M;
    while ~isempty(node),       % while root or above
        for j = jrange,
            children = node.children{j};
            node.b(j) = [children.f] * ( [children.r] + discount * max([children.b]) )';
            % [children.b] is a matrix with b-values of actions on
            % columns, so the max takes the maximum across actions and returns an L-length vector
            % that can be operated on in dimension sync with the rest
            node.util(j) = [children.f] * ( [children.r] + discount * max([children.util]) )';
        end;
        jrange = node.parentj;
        node = node.parent;
    end;
    
end;    % end main tree construction loop

% finally, return a greedy action in util at the tree root
[~, jmax] = max(root.util);
uroot = U(:, jmax);

% % save old tree if configured, and tree doesn't exceed maximum size
% if cfg.reusetree, 
%     if length(tree) <= cfg.maxtreesize, oldtree = tree; 
%     else oldtree = [];
%     end;
% end;

% output stats if requested
if nargout >= 2,
    planinfo.n = tau;
    planinfo.depth = max([tree.depth]);
    planinfo.trun = cputime - t;
    planinfo.treesize = length(tree);
    planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d, treesize=%d', ...
        tau, planinfo.depth, planinfo.treesize); 
end;

end % main function
